"use strict";
this.name           = "SC Core";
this.author         = "Lone_Wolf";
this.copyright      = "(C) 2011-2012, 2014-2015 Lone_Wolf.";
this.licence        = "CC-by-SA 4.0";
this.description    = "calculate shield values when cycling";
this.version        = "1.12";

    this._sc_cycler_loss =
    //  power loss for no cycler, basic,standard, advanced adjustments based on adjustment / 64
    //  64 is used because that's the amount of energy in 1 energybank.
      [
	[ 0, 0, 0, 0, 0, 0 ],     // no cycler present, so no power loss
	[ 4, 5, 6, 8, 12, 14 ],   // basic cycler
	[ 3, 4, 4, 6, 8, 10 ],    // standard cycler
	[ 2, 2, 3, 4, 6, 7 ],     // advanced cycler
      ];
    this._sc_manual_reduction =
    // powerloss multiplier for manual configurators
    // none, basic, standard, advanced
    [ 1, 1, 0.50, 0.20];

this.startUp = function()
  {
    if ( worldScripts["Shield Cycler"].startUp ) { worldScripts["Shield Cycler"].startUp(); };
  };

this._sc_sc_adjust = function(init)
  {
    // setup things for player.ship shield cycling
    var adjust_input =
      {
	caller: "Shield Cycler",
	setting: worldScripts["Shield Cycler"]._sc_settings.sc.current_configuration,
	functional: worldScripts["Shield Cycler"]._sc_settings.sc.functional,
	fwd: player.ship.forwardShield,
	fwd_threshold: ( player.ship.maxForwardShield * worldScripts["Shield Cycler"]._sc_settings.sc.threshold * 0.01 ),
	aft: player.ship.aftShield,
	aft_threshold: ( player.ship.maxAftShield * worldScripts["Shield Cycler"]._sc_settings.sc.threshold * 0.01 ),
	version: worldScripts["Shield Cycler"]._sc_settings.sc.version,
	manual_version: worldScripts["Shield Cycler"]._sc_settings.sc.manual_version,
	init: init,
	max_energy: player.ship.maxEnergy
      };
    var adjust_output = {power: 0, fwd:0, aft:0};
    adjust_output = this._sc_adjust(adjust_input);
    if ( adjust_input.init )
      { player.commsMessage( "Cycler configuration : " + worldScripts["Shield Cycler"]._sc_const.names[ adjust_input.setting ] ); };
    if ( adjust_output.power > 0 )
      {
        // apply changes to shields & energy
	player.ship.forwardShield = adjust_output.fwd;
	player.ship.aftShield = adjust_output.aft;
	player.ship.energy -= adjust_output.power;
      };
    return adjust_output.power;
  };

  
  
  
  
  
this._sc_adjust = function (adjust)
  {
    /* adjust is an object set by the calling function
      properties of the object adjust :
	name		type	meaning
     	caller		string	identifier of calling oxp
     	setting		int	0 = disabled, 1 = aft, 2 = forward, 3 = equal
	functional	%	indicates how much damage shieldcycler device has, 100 = no damage
	fwd		int	current forward shield strength
	fwd_threshold	int	see code
	aft		int	aft shield strength
	aft_threshold	int	see code
	version		int	number 0,1,2,3 determining which version of automatic shield cycler is present
	manual_version	string	number 0,1,2,3 determining which version of manual configurator is present
	init		bool	true : calc powerloss for switching to another setting
				false : calc powerloss for cycling
	max_energy	int	needed to calculate powerloss for init = true
	
      returns an object with these properties :
	power: amount of energy used for cycling or -1 in case of error
	fwd: new forward shield value
	aft: new aft shield value
	
	NOTE : this function is intended to be usable for both player & npc ships
    */
    // set result at power 0, fwd & aft at input values. This corresponds with no changes needed.
    var result = { power:0,fwd:adjust.fwd,aft:adjust.aft};
    var energy_transfer = -1;
    var desired_transfer = 0;
    switch ( worldScripts["Shield Cycler"]._sc_const.names[adjust.setting] )
    {
      case "Equal":
          if ( adjust.aft !== adjust.fwd )
            {
              if ( (adjust.fwd < adjust.fwd_threshold) || (adjust.aft < adjust.aft_threshold)  )
                {
                  desired_transfer = 0.5 * Math.abs( adjust.fwd - adjust.aft );
		  energy_transfer = desired_transfer * adjust.functional * 0.01 ;
                  result.fwd = ( adjust.fwd < adjust.aft ) ? adjust.fwd + energy_transfer : adjust.fwd - energy_transfer ;
                  result.aft = ( adjust.fwd < adjust.aft ) ? adjust.aft - energy_transfer : result.aft = adjust.aft + energy_transfer ;
                };
            };
          break;
      case "Forward":
          if ( adjust.fwd < adjust.fwd_threshold )
            { 
              desired_transfer  = ( desired_transfer <= adjust.aft ) ? adjust.fwd_threshold - adjust.fwd : adjust.aft ;
              energy_transfer = desired_transfer * adjust.functional * 0.01 ;
	      result.fwd = adjust.fwd + energy_transfer;
              result.aft = adjust.aft - energy_transfer;
            };
          break;
      case "Aft":
          if ( adjust.aft < adjust.aft_threshold )
            {
              desired_transfer  = ( desired_transfer > adjust.fwd ) ? adjust.aft_threshold - adjust.aft : adjust.fwd ;
	      energy_transfer = desired_transfer * adjust.functional * 0.01 ;
	      result.fwd = adjust.fwd - energy_transfer;
              result.aft = adjust.aft + energy_transfer;
            };
          break;
      case "Disabled":
          // no adjustments are done in disabled configuration, so no energy is transferred
          desired_transfer = 0;
	  break;
      default:
	// should never happen unless caller has set an invalid mode
	desired_transfer = -1;
  	result.power = -1;
	log(this.name,JSON.stringify(adjust));
	log(this.name,worldScripts["Shield Cycler"]._sc_const.names[adjust.setting]);
	log(this.name,JSON.stringify(result));
	log(this.name,"error while adjusting shields");
        break;
    };
    // no transfer done, so no power loss
    if ( desired_transfer = 0 ) { result.power = 0; return result;};
    // incase something went very wrong
    if ( desired_transfer < 0 ) { return result; };
    if ( adjust.init )
	    { 
	      result.power = 2 * ( Math.floor( adjust.max_energy / 64) + 1);
	    }
	  else
	    { 
	      result.power = this._sc_cycler_loss[adjust.version] [Math.floor(desired_transfer / 64)];
	      result.power *= this._sc_manual_reduction[adjust.manual_version];
	    };
    return result;
  };
