"use strict";
this.name           = "Shield Cycler";
this.author         = "Lone_Wolf";
this.copyright      = "(C) 2011-2012, 2014-2015 Lone_Wolf.";
this.licence        = "CC-by-SA 4.0";
this.description    = "Keeps shield energy aligned with a specific configuration";
this.version        = "1.12";

/* Main script for Shield Cycler
 * + functions used in other SC scripts
 * + functions used by/for other oxps
*/

/* All settings that need saving are stored in missionVariables.sc_settings using JSON
   for ease of use & performance the SC scripts internally use the object this._sc_settings / worldScripts["Shield Cycler"]._sc_settings
     
	name				Type	Introduced in	Used by / Comment
	_sc_settings			obj	1.10
	  sc				obj	1.10
		threshold		int	0.21.2		OxpConfig
								    percentage, can be 0-100
		start_configuration     int	0.21.2		OxpConfig
								    number , 0 = disabled 1 = aft 2 = fwd 3 = equal
		current_configuration	int	1.10		shipscript events, equipment events
								    number , 0 = disabled 1 = aft 2 = fwd 3 = equal
		cycler_mask            	int	1.10		OxpConfig
								    bitmask (size = 4 ) used to keep track which settings are available for cycling
								    bit order (MSB first) : equal, forward, aft , disabled
								    bit = 1 : available, 0 not available
		functional		int	1.10		equipment condition script
								    percentage, 100 is fully functional, 20 is minimum 
								    for 1.10 only 100 is used
								    in 1.11+ this value will be decreased when EQ_SC_SHIELD_CYCLER_INTERNAL or EQ_SC_MANUAL_CONFIGURATOR_INTERNAL get hit
		version			int	0.21.2		SC_Core.js
								 number, 0=none 1 = basic , 2 = standard, 3 =advanced 
		manual_version		int	0.21.2		SC_Core.js
								 number, 0=none 1 = basic , 2 = standard, 3 =advanced 
          general			obj     1.10
		disabled		bool	1.10		if true, all devices handled by this oxp will do nothing
		sc_disabled		bool	1.10		if true, presence of EQ_SC_SHIELD_CYCLER_INTERNAL & EQ_SC_MANUAL_CONFIGURATOR_INTERNAL will have no effect
		savegame_version	int	1.10		This is a special var intended to track missionVariables changes

  NOTE : missionvariables are loaded at startup, then converted into this._sc_setting object .
	 during game the missionvariables are NOT kept uptodate.
	 they are stored only when player saves game, done in SC_Worldscript_Events.js .

  equipment items
    EQ_SC_SHIELD_CYCLER_INTERNAL          - indicates any shield cycler/manual configurator is installed
    EQ_SC_SHIELD_CYCLER_BASIC
    EQ_SC_SHIELD_CYCLER_STANDARD
    EQ_SC_SHIELD_CYCLER_ADVANCED
    EQ_SC_MANUAL_CONFIGURATOR_INTERNAL    - present if a manual configurator is installed
    EQ_SC_MANUAL_CONFIGURATOR_BASIC
    EQ_SC_MANUAL_CONFIGURATOR_STANDARD
    EQ_SC_MANUAL_CONFIGURATOR_ADVANCED
    EQ_SC_SELL_SHIELD_CYCLER
    EQ_SC_SELL_MANUAL
    EQ_SC_SC_REPAIR
    
    NOTE EQ_SC_SHIELD_CYCLER_INTERNAL and EQ_SC_MANUAL_CONFIGURATOR_INTERNAL will be present in the savefile but NOT visible on the F5 equipment screen

*/

// RO stuff, i haven't been able to figure out how to make these read-only, so the values here can be changed by code, but that's not supposed to happen
this._sc_const =
  {
    none:0,basic:1,standard:2,advanced:3,
    cycle: { disabled:0,aft:1,forward:2,both:3,count:4 },
    names: [ "Disabled","Aft","Forward","Equal"],
    savegame_version:0
  };

// WORM stuff
this._sc_worm =
  {
    sc_cost: [0, 0, 0, 0 ],
    manual_cost: [ 0, 0, 0, 0 ]
  };
  
// R/W stuff
this._sc_settings =
  {
    sc:
      {
	threshold:100,
	start_configuration:this._sc_const.both,
	current_configuration:this._sc_const.both,
	cycler_mask:0x00f,
	functional:100,
	version:this._sc_const.none,
	manual_version:this._sc_const.none
      },
    general: { disabled:false,sc_disabled:false,savegame_version:this._sc_const.savegame_version }
  };
	
// Properties for internal use
this._sc_error = 0;
    /* used for error tracking
    * 0 = no error
    * 1 = savegame to old
    * 2 = savegame to new
    */

// for OxpConfig
// these need to be kept in sync at all times 
this._sc_oxp_status = this._sc_settings.general.disabled;
this._sc_sc_status = this._sc_settings.general.sc_disabled;
// only changed through oxpconfig, synced from this.oxpcNotifyOnChange()
this._sc_threshold = this._sc_settings.sc.threshold;
this._sc_start_configuration = this._sc_settings.sc.start_configuration;
this._sc_cycler_mask = this._sc_settings.sc.cycler_mask;
this.deactivated = false;
    /*  this property is used by OxpConfig to detect whether it should allow configuration changes.
	deactivated = false : default behaviour
	deactivated = true :  OxpConfig will not allow configuration changes
    */

this.oxpcSettings = 
  { 
    Info:
      {
        Name:"Shield Cycler",DynamicNotify:true,LeaveData:true,
        InfoB:"enable / disable Shield Cycler oxp functionality / the devices it manages",
        InfoS:"Threshold % determines when adjusting is done. 100 = always, 0 = never. Start configuration is the configuration that will be set upon launch (3 = equal, 2 = fwd, 1= aft, 0 = disabled).",
        InfoE:"enable / disable the availability of the settings for manual cycling. Note : if you disable all, Shield Cycler oxp will enable the disabled setting"
      },
    Bool0: {Name:"_sc_oxp_status",Def:false,Hide:false,Desc:"disable SC OXP."},
    Bool1: {Name:"_sc_sc_status",Def:false,Hide:false,Desc:"disable cycler devices."},
    SInt0: {Name:"_sc_threshold",Def:100,Max:100,Hide:true,Desc:"threshold %"},
    SInt1: {Name:"_sc_start_configuration",Def:3,Max:3,Hide:true,Desc:"launch setting"},
    EInt0: {Name:"_sc_cycler_mask",Def:0x001,Max:0x00f,Hide:true,Desc:["Disabled","Aft","Forward","Equal"]}
  };

this.startUp = function()
  {
    // CCL & OxpConfig are required for SC, but this is taken care of by Oolite itself through the manifest.plist file.
    // This version of SC requires 1.81+ , that is checked through manifest.plist & requires.plist
    switch ( worldScripts["SC check savegame compatiblity"]._sc_check_savegame(this._sc_settings.general.savegame_version) )
      {
	case 0:
	  // savegame to old
	  this._sc_settings.general.disabled = true;
	  this.deactivated= true;
	  this._sc_error = 1;
	  player.commsMessage("Your savegame is to old for this version of Shield Cycler",10);
	  player.commsMessage("Shield Cycler is DISABLING itself now",10);
	  player.commsMessage("ask for help on Oolite BBS.",10);
	  log(this.name," - DISABLED because savefile is to old");
	  break;
	case 1:
	  // savegame needs converting from old-style missionvars to JSON
	  worldScripts["SC savegame conversion"]._sc_convert_0_21_2();
	  this._sc_update_oxpconfig_vars();
	  log(this.name," - savegame missionvars converted into _sc_settings");
	  break;
	case 2:
	  // missionvars not present, will be created when game is saved
	  break;
	case 3:
	  // everything in order, load missionvars
	  this._sc_settings = JSON.parse(missionVariables.sc_settings);
	  this._sc_update_oxpconfig_vars();
  	  if ( ( this._sc_settings.sc.version > this._sc_const.basic ) ) { this._sc_change_configuration(1, false); };
	  if ( this._sc_settings.sc.version == this._sc_const.advanced ) { this._sc_change_configuration(2, false); };
	  if ( this._sc_settings.sc.manual_version == this._sc_const.advanced ) { this._sc_change_configuration(3, false); };	  
	  log(this.name," - savegame missionvars loaded");
	  break;
	case 4:
	  // savegame to new
	  this._sc_settings.general.disabled = true;
	  this.deactivated= true;
	  this._sc_error = 2;
	  player.commsMessage("Your savegame is to NEW for this version of Shield Cycler",10);
	  player.commsMessage("Shield Cycler is DISABLED",10);
	  player.commsMessage("ask for help on Oolite BBS.",10);
	  log(this.name," - DISABLED because savefile  is to NEW");
	  break;
	default:
	  break;
      };
    if ( !this._sc_settings.general.disabled )
      {
        // calculate total price values
        this._sc_worm.sc_cost[1] = EquipmentInfo.infoForKey("EQ_SC_SHIELD_CYCLER_BASIC").price;
        this._sc_worm.sc_cost[2] = this._sc_worm.sc_cost[1] + EquipmentInfo.infoForKey("EQ_SC_SHIELD_CYCLER_STANDARD").price;
        this._sc_worm.sc_cost[3] = this._sc_worm.sc_cost[2] + EquipmentInfo.infoForKey("EQ_SC_SHIELD_CYCLER_ADVANCED").price;
        this._sc_worm.manual_cost[1] = EquipmentInfo.infoForKey("EQ_SC_MANUAL_CONFIGURATOR_BASIC").price;
        this._sc_worm.manual_cost[2] = this._sc_worm.manual_cost[1] + EquipmentInfo.infoForKey("EQ_SC_MANUAL_CONFIGURATOR_STANDARD").price;
        this._sc_worm.manual_cost[3] = this._sc_worm.manual_cost[2] + EquipmentInfo.infoForKey("EQ_SC_MANUAL_CONFIGURATOR_ADVANCED").price;
      };
        
  delete this.startUp;
  };

// Helper functions for all SC scripts including this one
  this._sc_change_configuration = function(setting, hide)
  {
    // sets/resets visibility of configure options in OCPConfig
    switch ( setting )
      {
        case 1:
            this.oxpcSettings.SInt0.Hide = hide;
            break;
        case 2:
            this.oxpcSettings.SInt1.Hide = hide;
            break;
        case 3:
            this.oxpcSettings.EInt0.Hide = hide;
            break;
        default:
            break;
      };
  };
  
this._sc_update_oxpconfig_vars = function()
  {
    this._sc_oxp_status = this._sc_settings.general.disabled;
    this._sc_sc_status = this._sc_settings.general.sc_disabled;
    this._sc_threshold = this._sc_settings.sc.threshold;
    this._sc_start_configuration = this._sc_settings.sc.start_configuration;
    this._sc_cycler_mask = this._sc_settings.sc.cycler_mask;  
  };

this._sc_remove_manual = function()
  {
    // remove manual configurator devices only
    // returns refund value in credits
    var refund1 = this._sc_worm.manual_cost[this._sc_settings.sc.manual_version];
    refund1 = refund1 * this._sc_settings.sc.functional * 0.01;
    player.ship.removeEquipment("EQ_SC_MANUAL_CONFIGURATOR_INTERNAL");
    this._sc_settings.sc.manual_version  = this._sc_const.none;
    this._sc_settings.sc.cycler_mask  = 0x00f;
    this._sc_update_oxpconfig_vars();
    refund1 = parseInt(refund1 * 0.1 * 0.6);
    return refund1;
  };  

this._sc_remove_shield_cycler = function()
  {
    // remove all shield cycler devices
    // returns refund value in credits
    var refund1 = 0;
    // if manual configurator devices are present, remove them first
    if ( this._sc_settings.sc.manual_version != this._sc_const.none )
      {
        refund1 = this._sc_remove_manual();
      }
    refund1 = refund1 + this._sc_worm.sc_cost[ this._sc_settings.sc.version ];
    switch ( this._sc_settings.sc.version )
      {
        case this._sc_const.advanced:
           this._sc_change_configuration(2, true);
        case this._sc_const.standard:
           this._sc_change_configuration(1, true);
           break;
        default:
          break;
      };
    refund1 = refund1 * this._sc_settings.sc.functional * 0.01;
    player.ship.removeEquipment("EQ_SC_SHIELD_CYCLER_INTERNAL");
    this._sc_settings.sc.version  = this._sc_const.none;
    this._sc_settings.sc.start_configuration = this._sc_const.both;
    this._sc_settings.sc.threshold = 100;
    this._sc_update_oxpconfig_vars();
    this._sc_settings.sc.functional = 100;
    refund1 = parseInt(refund1 * 0.1 * 0.6);
    return refund1;
  };
  

  
// functions for use by other oxps

this._sc_stop = function(device)
  {
/*	device is a string to indicate which devices handled by SC oxp need to be stopped
	currently has 2 possible values : 
	"ALL" : disable all devices
	"SC" : disables just the shieldcycler
	In future versions there will be more options

	returns
	-3 if SC was disabled
	-2 if device was stopped already
	-1 device string invalid
	0 if stopping was succesfull
	NOTE : no messages to player or logfile, imo this is the responsibility of the calling oxp
	NOTE : calling worldScripts["Shield Cycler"]._sc_stop("SC") DOES NOT disable the entire oxp.
*/
    var success = -3;
    if ( this._sc_error == 0 )
      {
	switch ( device )
	  {
	    case "ALL":
	      if ( this._sc_settings.general.disabled )
		{ success = -2; }
	      else 
		{ 
		  this._sc_settings.general.disabled = true;
		  this._sc_update_oxpconfig_vars();
		  success = 0;
		};
	      break;
	    case "SC":
	      if ( this._sc_settings.general.sc_disabled )
		{ success = -2; }
	      else
		{
		  this._sc_settings.general.sc_disabled = true;
		  this._sc_update_oxpconfig_vars;
		  success = 0;
		};
	      break;
	    default:
	      succes = -1;
	      break;
	  };
      };
    return success;
  };

this._sc_start = function(device)
  {
/*	device is a string to indicate which devices handled by SC oxp need to be started
	currently has 2 possible values : 
	"ALL" : start all devices
	"SC" : start just the shieldcycler

	returns
	-3 if SC is disabled
	-2 if device was started already
	-1 device string invalid
	0 if starting was succesfull
	Note : no messages to player or logfile, imo this is the responsibility of the calling oxp
*/
    var success = -3;
    var adjust_error = 0;
    if ( this._sc_error == 0 )
      {
	switch ( device )
	  {
	    case "ALL":
	      if ( !this._sc_settings.general.disabled )
		{ success = -2; }
	      else
		{
		  this._sc_settings.general.disabled = false;
		  this._sc_settings.general.sc_disabled = false;
		  this._sc_update_oxpconfig_vars();
		  // start sc_device
		  if ( this._sc_settings.sc.version > this._sc_const.none ) 
		    { 
		      adjust_error = worldScripts["SC Core"]._sc_sc_adjust(true);
		      if ( adjust_error < 0 )
			{
			  player.commsMessage("ERROR while starting shield cycler device, check latest log",10); 
			  log(this.name," unable to adjust shields while starting entire oxp");
			};
		    };
		  success = 0;
		};
	      break;
	    case "SC":
	      if ( !this._sc_settings.general.sc_disabled )
		{ success = -2; }
	      else
		{
		  this._sc_settings.general.sc_disabled = false;
		  this._sc_update_oxpconfig_vars();
		  if ( this._sc_settings.sc.version > this._sc_const.none ) 
		    { 
		      adjust_error = worldScripts["SC Core"]._sc_sc_adjust(true);
		      if ( adjust_error  < 0 )
			{
			  player.commsMessage("ERROR while starting shield cycler device, check latest log",10); 
			  log(this.name," unable to adjust shields while starting sc device");
			};
		      success = 0;
		    };
		};
	      break;
	    default:
	      success = -1;
	      break;
	  };
      };
    return success;
  };


this._sc_store_devices = function()
  {
    // Shield Cycler devices are not portable between ships and will stay that way.
    // This poses a problem for oxps (like Ship Storage Helper) that store ships for later retrieval
    // returns an object with 2 strings :
    // first holds SC values in JSON format
    // second is encrypted version of first
    var sc_values =
      {
        json: "",
        enc: ""
      };
    sc_values.json = JSON.stringify(this._sc_settings);
    var helper = new worldScripts.Cabal_Common_Functions.Cabal_Common();
    sc_values.enc = helper.strEncrypt( sc_values.json, "ShieldCycler" );
    // since ship is stored, ignore refund
    var temp = this._sc_remove_shield_cycler();
    return sc_values;
  }
  
this._sc_retrieve_devices = function(sc_values)
  {
    // counterpart of this._sc_store_devices
    // input : an object created by this._sc_store_devices
    // returns a code value :
    // 0 : import sucessfull
    // 1 : input != an object
    // 2 : object has wrong structure
    // 3 : invalid json data
    // 4 : plain version doesn't match encrypted version
    // 5 : stored data to old
    // 6 : stored data to new
    
    if ( typeof sc_values !== "object" ) { return 1; };
    if ( typeof sc_values.json !== "string" || sc_values.enc !== "string" ) { return 2; };
    var temp = "";
    try
      { temp = JSON.parse(sc_values.json) }
    catch(e)
      { return 3; };
    var helper = new worldScripts.Cabal_Common_Functions.Cabal_Common();
    if ( sc_values.json !== helper.strDecrypt( sc_values.json, "ShieldCycler" ) ) { return 4; };
    // Data is valid JSON and created by SC
    var version_diff = parseInt( this._sc_settings.general.savegame_version - sc_values.general.savegame_version  ) ;
    if ( version_diff < -1 ) { return 5;} ; // to old
    if ( version_diff > 1 ) { return 6; } ; // to new
    // version_diff == 1 is impossible for now, but in future versions it will indicate conversion is needed
    // only case left is when version_diff == 0
    this._sc_settings = temp;
    this._sc_update_oxpconfig_vars();
    if ( ( this._sc_settings.sc.version > this._sc_const.basic ) ) { this._sc_change_configuration(1, false); };
    if ( this._sc_settings.sc.version == this._sc_const.advanced ) { this._sc_change_configuration(2, false); };
    if ( this._sc_settings.sc.manual_version == this._sc_const.advanced ) { this._sc_change_configuration(3, false); };     
    log(this.name," - imported stored ship data");
    return 0;
  };

// OXPConfig2 related

this.oxpcNotifyOnChange = function(n)
  {
    // called by OXPConfig when player changes values
    switch (n)
      {
	case 1:
	  // booleans changed
	  var error = 0;
	  if ( this._sc_oxp_status !== this._sc_settings.general.disabled )
	    {
	      var old_status = this._sc_settings.general.disabled;
	      error = ( this._sc_oxp_status ) ? this._sc_stop("ALL") : this._sc_start("ALL");
	      if ( error == 0 )
		{ 
		  this._sc_settings.general.disabled = this._sc_oxp_status;
		}
	      else
		{ 
		  log(this.name," old value:" + old_status);
		  log(this.name," new value:" + this._sc_oxp_status);
		  log(this.name," - problem with applying change for SC oxp status");
		  this._sc_oxp_status = old_status;
		};
	    }
	  else
	    {
	      if ( this._sc_sc_status !== this._sc_settings.general.sc_disabled )
		{
		  if ( this._sc_settings.sc.version > this._sc_const.none )
		    {
		      this._sc_sc_status = this._sc_settings.general.sc_disabled;
		      log(this.name," Buy a shieldcycler before you try to change it's status !");
		    }
		  else
		    {
		      old_status = this._sc_settings.general.sc_disabled;
		      if ( this._sc_sc_status ) { error = this._sc_stop("SC"); } else { error = this._sc_start("SC"); }
		      if ( error == 0 )
			{ 
			  this._sc_settings.sc.general.sc_disabled = this._sc_sc_status;
			}
		      else
			{ 
			  log(this.name," old value:" + this._sc_settings.general.sc_disabled );
			  log(this.name," new value:" + this._sc_sc_status );
			  log(this.name," - problem with applying change for SC device status");
			  this._sc_sc_status = old_status;
			};
		    };
		};
	    };
	  break;
	case 2:
	  // integers changed
	  this._sc_settings.sc.threshold = this._sc_threshold;
	  this._sc_settings.sc.start_configuration = this._sc_start_configuration;
	  break;
	case 4:
	  // bitmask changed
	  if ( this._sc_cycler_mask == 0x000 ) { this._sc_cycler_mask = 0x001;}
	  this._sc_settings.sc.cycler_mask = this._sc_cycler_mask;
	  break;
	default:
	  break;
      };
  };
